function [MIh,score_cv] = mLSMI(x,y,y_type,sigma_list,lambda_list,b)
%
% Least-Squares Mutual Information with Multiplicative Kernel Models (with likelihood cross validation)
%
% Estimating a squared-loss variant of mutual information 
%    \frac{1}{2}\int\int (\frac{p_{xy}(x,y)}{p_x(x)p_y(y)}-1)^2 p_x(x)p_y(y) dx dy
% from input-output samples
%    { (x_i,y_i) | x_i\in R^{dx}, y_i\in R^{dy} }_{i=1}^n 
% drawn independently from a joint density p_{xy}(x,y).
% p_x(x) and p_y(y) are marginal densities of x and y, respectively.
% 
% Usage:
%       [MIh,score_cv]=mLSMI(x,y,y_type,sigma_list,lambda_list,b)
%
% Input:
%    x          : dx by n input sample matrix
%    y          : If y_type is 0, dy by n output sample matrix.
%                 If y_type is 1, 1 by n label vector.
%    y_type     : If y_type is 0, Gaussian kernel is used.
%                 If y_type is 1, delta kernel is used.
%    sigma_list : (candidates of) Gaussian width
%                 If sigma_list is a vector, one of them is selected by cross validation.
%                 If sigma_list is a scalar, this value is used without cross validation
%                 If sigma_list is empty/undefined, Gaussian width is chosen from
%                 some default canditate list by cross validation
%    lambda_list: (OPTIONAL) regularization parameter
%                 If lambda_list is a vector, one of them is selected by cross validation.
%                 If lambda_list is a scalar, this value is used without cross validation
%                 If lambda_list is empty, Gaussian width is chosen from
%                 some default canditate list by cross validation
%    b          : number of Gaussian centers (if empty, b=50 is used);
%
% Output:
%    MIh        : estimated mutual information between x and y
%    score_cv   : cross validation score
%
% (c) Tomoya Sakai and Masashi Sugiyama, Department of Compter Science, Tokyo Institute of Technology, Japan.
%     sakai@sg.cs.titech.ac.jp,
%     sugi@cs.titech.ac.jp

if nargin < 3
    error('number of input arguments is not enough!');
end

n = size(x,2);
ny = size(y,2);

if n ~= ny
    error('x and y must have the same number of samples!');
end

if nargin < 4 || isempty(sigma_list)
    sigma_list = logspace(-2,2,9);
end
if nargin < 5 || isempty(lambda_list)
    lambda_list = logspace(-3,1,9);
end

if nargin < 6 || isempty(b)
  b = 20;
end

b=min(n,b);

center_index=randperm(n);
center_index=center_index(1:b);
center_x=x(:,center_index);
center_y=y(:,center_index);

clear center_index

dist2_x=bsxfun(@plus,sum(x.^2,1),bsxfun(@minus,sum(center_x.^2,1)',2*center_x'*x));
if y_type==0
    dist2_y=bsxfun(@plus,sum(y.^2,1),bsxfun(@minus,sum(center_y.^2,1)',2*center_y'*y));
else
    dist2_y=double(bsxfun(@eq,y,center_y'));
end

clear x y center_x center_y

if length(sigma_list)==1 && length(lambda_list)==1
  sigma=sigma_list;
  lambda=lambda_list;
  score_cv=-inf;
else
  %%%%%%%%%%%%%%%% Searching Gaussian kernel width `sigma_chosen'
  %%%%%%%%%%%%%%%% and regularization parameter `lambda_chosen' 
  fold=5;
  fold_index=[1:fold];
  tmp=floor([0:n-1]*fold./n)+1;
  cv_index=tmp(randperm(n));
  scores_cv=zeros(length(sigma_list),length(lambda_list));
  n_cv=zeros(1,fold); n_te=zeros(1,fold);

  for i=fold_index
    n_cv(i)=sum(cv_index~=i);
    n_te(i)=sum(cv_index==i);
  end
  
  for sigma_index=1:length(sigma_list)
    Phix_sigma=exp(-dist2_x/(2*sigma_list(sigma_index)^2));
    if y_type==0
        Phiy_sigma=exp(-dist2_y/(2*sigma_list(sigma_index)^2));
    else
        Phiy_sigma=dist2_y;        
    end
    for i=fold_index
      [S,M]=schur(Phix_sigma(:,cv_index~=i)*Phix_sigma(:,cv_index~=i)'/n_cv(i));
      [T,N]=schur(Phiy_sigma(:,cv_index~=i)*Phiy_sigma(:,cv_index~=i)'/n_cv(i));
      HTilde_cv=Phix_sigma(:,cv_index~=i)*Phiy_sigma(:,cv_index~=i)'/n_cv(i);
      KTilde_te=Phix_sigma(:,cv_index==i)*Phix_sigma(:,cv_index==i)'/n_te(i);
      LTilde_te=Phiy_sigma(:,cv_index==i)*Phiy_sigma(:,cv_index==i)'/n_te(i);
      HTilde_te=Phix_sigma(:,cv_index==i)*Phiy_sigma(:,cv_index==i)'/n_te(i);
      for lambda_index=1:length(lambda_list)
        lambda=lambda_list(lambda_index);
        Thetah_cv=S*bsxfun(@rdivide,S'*HTilde_cv*T,diag(M)*diag(N)'+lambda)*T';
        wh_cv=trace(Thetah_cv'*KTilde_te*Thetah_cv*LTilde_te)/2-trace(Thetah_cv'*HTilde_te);
        scores_cv(sigma_index,lambda_index)=scores_cv(sigma_index,lambda_index)+wh_cv/fold;
      end % fold
    end % lambda
  end % sigma
  [scores_cv_tmp,lambda_chosen_index]=min(scores_cv,[],2);
  [score_cv,sigma_chosen_index]=min(scores_cv_tmp);
  lambda=lambda_list(lambda_chosen_index(sigma_chosen_index));
  sigma=sigma_list(sigma_chosen_index);
end %length(sigma_list)==1 && length(lambda_list)==1

%%%%%%%%%%%%%%%% Computing the final solution `MIh'
Phix_sigma=exp(-dist2_x/(2*sigma^2));
if y_type==0
    Phiy_sigma=exp(-dist2_y/(2*sigma^2));
else
    Phiy_sigma=dist2_y;
end
KTilde=Phix_sigma*Phix_sigma'/n;
LTilde=Phiy_sigma*Phiy_sigma'/n;
HTilde=Phix_sigma*Phiy_sigma'/n;
[S,M]=schur(KTilde);
[T,N]=schur(LTilde);
Thetah=S*bsxfun(@rdivide,S'*HTilde*T,diag(M)*diag(N)'+lambda)*T';

%MIh=trace(Thetah'*HTilde)/2-1/2;
MIh=-trace(Thetah'*KTilde*Thetah*LTilde)/2+trace(Thetah'*HTilde)-1/2;
